/*
** Copyright (C) 1999 Christian Werner
**
** This program and library is free software; you can redistribute it and/or
** modify it under the terms of the GNU (Library) General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU (Library) General Public License for more details.
**
** You should have received a copy of the GNU (Library) General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package swishej;

import java.util.*;

/**
 * Representation of search result lists
 * @see swishej.Search
 */

class ResultList {
    /**
     * First result in list.
     */
    Result first;

    /**
     * Last result in list.
     */
    Result last;

    /**
     * Create new empty result list.
     */
    ResultList() {
        first = null;
	last = null;
    }

    /**
     * Merge result described by three integer numbers into
     * a local result list
     * @param f file number (short)
     * @param r rank for sorting
     * @param s structure tag
     * @param list table which holds the merged results
     */
    static void mergeResult(int f, int r, int s, Hashtable list) {
        Integer key = new Integer(f);
	Result res = (Result) list.get(key);
	if (res == null) {
	    res = new Result(f, r, s);
	    list.put(key, (Object) res);
	} else {
	    res.rank += r;
	    res.structure |= s;
	}
    }

    /**
     * Add result described by three integer numbers to
     * result list
     * @param f file number (short)
     * @param r rank for sorting
     * @param s structure tag
     */
    void add(int filenum, int rank, int structure) {
        Result result = new Result(filenum, rank, structure);
	if (last == null) {
	    first = result;
	} else {
	    last.next = result;
	}
	last = result;
    }

    /**
     * Test result list being empty.
     * @return true if the result list is empty, false otherwise
     */
    boolean empty() {
        return first == null;
    }

    /**
     * OR two result lists by merging all results and
     * forming a new result list.
     * @param r1 first result list or null
     * @param r2 second result list or null
     * @return new ResultList
     */
    static ResultList orresultlists(ResultList r1, ResultList r2) {
        if (r1 == null) {
	    return r2;
	}
	if (r2 == null) {
	    return r1;
	}
	Hashtable list = new Hashtable();
	Result r = r1.first;
	while (r != null) {
	    mergeResult(r.filenum, r.rank, r.structure, list);
	    r = r.next;
	}
	r = r2.first;
	while (r != null) {
	    mergeResult(r.filenum, r.rank, r.structure, list);
	    r = r.next;
	}
	ResultList rl = new ResultList();
	Enumeration e = list.elements();
	while (e.hasMoreElements()) {
	    r = (Result) e.nextElement();
	    rl.add(r.filenum, r.rank, r.structure);
	}
	return rl;
    }

    /**
     * AND two result lists by traversing over all elements
     * and accumulating rank information for elements which
     * occur in both lists. A new result list is produced.
     * @param r1 first result list or null
     * @param r2 second result list or null
     * @return new ResultList
     */
    static ResultList andresultlists(ResultList r1, ResultList r2,
				    int andLevel) {
        ResultList newl = new ResultList();
	if (r1 == null || r2 == null) {
	    return newl;
	}
	if (andLevel < 1) {
	    andLevel = 1;
	}
	Result r = r1.first;
	while (r != null) {
	    Result s = r2.first;
	    while (s != null) {
	        if (r.filenum == s.filenum) {
		    int newRank =
		        ((r.rank * andLevel) + s.rank) / (andLevel + 1);
		    newl.add(r.filenum, newRank, r.structure & s.structure);
		}
		s = s.next;
	    }
	    r = r.next;
	}
	return newl;
    }

    /**
     * NOT a result list and return a new result list.
     * @param rl result list to be NOTed
     * @param maxfilenum highest short file number
     * @return new ResultList
     */
    static ResultList notresultlist(ResultList rl, int maxfilenum) {
	 ResultList newl = new ResultList();
	 if (rl == null) {
	     return newl;
	 }
         Mark m = new Mark();
	 Result r = rl.first;
	 while (r != null) {
	     m.marknum(r.filenum);
	     r = r.next;
	 }
	 for (int i = 1; i <= maxfilenum; i++) {
	     if (!m.ismarked(i)) {
	         newl.add(i, 1000, Consts.IN_ALL);
	     }
	 }
	 return newl;
    }

}
