/*
** Copyright (C) 1999-2001 Christian Werner
** modified by InterRed GmbH 03/2003; getSortResult() added, expandstars() modified
**
** This program and library is free software; you can redistribute it and/or
** modify it under the terms of the GNU (Library) General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU (Library) General Public License for more details.
**
** You should have received a copy of the GNU (Library) General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package swishej;

import java.io.*;
import java.net.*;

/**
 * Class implementing the SWISH-E search.
 * @see swishej.SortResult
 * @see swishej.ResultList
 */

class SearchImpl {
    /**
     * Return rule constant given word.
     * @param word word to be checked.
     * @return rule constant, e.g. AND_RULE, OR_RULE etc.
     */
    static int getrulenum(String word) {
        if (word.compareTo("and") == 0) {
	    return Consts.AND_RULE;
	}
	if (word.compareTo("or") == 0) {
	    return Consts.OR_RULE;
	}
	if (word.compareTo("not") == 0) {
	    return Consts.NOT_RULE;
	}
	return Consts.NO_RULE;
    }

    /**
     * Check if given word is a rule.
     * @param word word to be checked.
     * @return true if word is a rule, e.g. "and", "or" etc.
     */
    static boolean isrule(String word) {
        int rule = getrulenum(word);
	return rule == Consts.AND_RULE ||
	    rule == Consts.OR_RULE ||
	    rule == Consts.NOT_RULE;
    }

    /**
     * Check if given word is a boolean rule.
     * @param word word to be checked.
     * @return true if word is a binary rule, i.e. "and" or "or"
     */
    static boolean isbooleanrule(String word) {
        int rule = getrulenum(word);
	return rule == Consts.AND_RULE ||
	    rule == Consts.OR_RULE;
    }

    /**
     * Check if given word is a unary rule.
     * @param word word to be checked.
     * @return true if word is a unary rule, i.e. "not"
     */
    static boolean isunaryrule(String word) {
        int rule = getrulenum(word);
	return rule == Consts.NOT_RULE;
    }

    /**
     * Given keyword string break it in words represented by a StrList.
     * @param words keyword string
     * @return StrList of keywords
     */
    static StrList mkwlist(String words) {
        StrList wl = new StrList();
	StringBuffer word = new StringBuffer();
	for (int i = 0; i < words.length(); i++) {
	    char c = words.charAt(i);
	    if (Character.isSpace(c) || c == '(' || c == ')' || c == '=') {
	        if (c == '=') {
		    if (word.length() > 0) {
		        if (i > 0 && words.charAt(i - 1) != '\\') {
			    wl.add(word.toString());
			    word.setLength(0);
			    wl.add("=");
			} else {
			    word.setLength(word.length() - 1);
			    word.append(IndexFile.toLowerCase(c));
			}
		    }
		} else {
		    if (word.length() > 0) {
		        wl.add(word.toString());
			word.setLength(0);
		    }
		    if (c == '(') {
		        wl.add("(");
		    } else if (c == ')') {
		        wl.add(")");
		    }
		}
	    } else {
	        word.append(IndexFile.toLowerCase(c));
	    }
	}
	if (word.length() > 0) {
	    wl.add(word.toString());
	}
	return wl;
    }

    /**
     * Check if given word is a meta word for the search.
     * @param word word to be checked
     * @return true if word is meta word
     */
    static boolean ismeta(String word) {
        if (word == null) {
	    return false;
	}
	return word.compareTo("=") == 0;
    }

    /**
     * Traverse StrList of keywords and reformat it for later parsing.
     * @param wl original keyword list
     * @return reformatted keyword list
     */
    static StrList fixnot(StrList wl) {
        int openparen = 0;
	int openmeta = 0;
	boolean hasparen = false;
	boolean hasmeta = false;
	boolean hasnot = false;
        StrList newwl = new StrList();
	String word = wl.first();
	while (word != null) {
	    char c = word.charAt(0);
	    if (c == '(' && hasnot) {
	        openparen++;
	    } else if (c == '(' && hasmeta) {
	        openmeta++;
	    } else if (c == ')' && hasnot) {
	        openparen--;
	    } else if (c == ')' && hasmeta) {
	        openmeta--;
	    }
	    if (ismeta(wl.succ())) {
	        hasmeta = true;
		newwl.add("(");
		newwl.add(word);
		newwl.add("=");
		word = wl.next();
		word = wl.next();
		continue;
	    }
	    if (word.compareTo("not") == 0) {
	        hasnot = true;
		newwl.add("(");
	    } else if (hasnot && openparen == 0) {
	        hasnot = false;
		newwl.add(word);
		newwl.add(")");
		word = wl.next();
		continue;
	    } else if (hasmeta && openmeta != 0) {
	        hasmeta = false;
		newwl.add(word);
		newwl.add(")");
		word = wl.next();
		continue;
	    }
	    newwl.add(word);
	    if (word.compareTo("=") == 0) {
	        hasmeta = true;
		newwl.add("(");
	    }
	    word = wl.next();
	}
	return newwl;
    }

    /**
     * Parse next term in keyword list and search accordingly.
     * @param wordlist keyword list
     * @param parseone if true only one word is considered
     * @param meta active meta word number for the search
     * @param inf index to be searched
     * @return ResultList of the term
     */
    static ResultList parseterm(StrList wordlist, boolean parseone, int meta,
				IndexFile inf) {
        ResultList rl = null;
	int andLevel = 0;
	int rulenum = Consts.OR_RULE;
	while (!wordlist.empty()) {
	    String word = wordlist.first();
	    if (rulenum == Consts.NO_RULE) {
	        rulenum = Consts.DEFAULT_RULE;
	    }
	    if (isunaryrule(word)) {
	        wordlist.rest();
	        rl = parseterm(wordlist, true, meta, inf);
		rl = ResultList.notresultlist(rl, inf.maxfilenum);
		rulenum = Consts.NO_RULE;
		continue;
	    }
	    if (isbooleanrule(word)) {
	        rulenum = getrulenum(word);
		wordlist.rest();
		continue;
	    }
	    if (rulenum == Consts.AND_RULE) {
	        andLevel++;
	    } else {
	        andLevel = 0;
	    }
	    if (word.charAt(0) == '(') {
		wordlist.rest();
		ResultList newrl = parseterm(wordlist, false, meta, inf);
		if (rulenum == Consts.AND_RULE) {
		    rl = ResultList.andresultlists(rl, newrl, andLevel);
		} else if (rulenum == Consts.OR_RULE) {
		    rl = ResultList.orresultlists(rl, newrl);
		}
		if (wordlist.empty()) {
		    break;
		}
		rulenum = Consts.NO_RULE;
		continue;
	    } else if (word.charAt(0) == ')') {
	        wordlist.rest();
		break;
	    }

	    String nextword = wordlist.first();
	    nextword = wordlist.next();
	    if (ismeta(nextword)) {
	        meta = inf.mh.getmeta(word);
		wordlist.rest();
		wordlist.rest();
		if (meta != 1) {
		    ResultList newrl = parseterm(wordlist, true, meta, inf);
		    if (rulenum == Consts.AND_RULE) {
		        rl = ResultList.andresultlists(rl, newrl, andLevel);
		    } else if (rulenum == Consts.OR_RULE) {
		        rl = ResultList.orresultlists(rl, newrl);
		    }
		    if (wordlist.empty()) {
		        break;
		    }
		    rulenum = Consts.NO_RULE;
		    meta = 1;
		    continue;
		}
	    }
	    rl = operate(rl, rulenum, wordlist.first(), meta, andLevel, inf);

	    wordlist.rest();
	    if (parseone) {
		break;
	    }
	    rulenum = Consts.NO_RULE;
	}
	return rl;
    }

    /**
     * Search the indexfile and accumulate results.
     * @param rl ResultList from earlier invokations
     * @param rulenum rule for merging results
     * @param word current keyword to be searched
     * @param meta active meta word number for the search
     * @param andLevel nesting level of "and" term
     * @param inf index to be searched
     * @return new ResultList
     */
    static ResultList operate(ResultList rl, int rulenum, String word,
			      int meta, int andLevel, IndexFile inf) {
        if (inf.stemming) {
	    /* apply stemming algorithm to the search term */
	    if (Consts.DEBUG) {
	        System.out.print("Stemmer: " + word + " -> ");
	    }
	    word = Stemmer.stem(word);
	    if (Consts.DEBUG) {
	        System.out.println(word);
	    }
	}
	if (inf.sh.isstopword(word) && !isrule(word)) {
	    if (rulenum == Consts.OR_RULE) {
	        return rl;
	    }
	}
	if (!inf.findWord(word)) {
	    if (rulenum == Consts.AND_RULE) {
	        return null;
	    }
	    if (rulenum == Consts.OR_RULE) {
	        return rl;
	    }
	}
	ResultList newrl = null;
        int index = word.indexOf("*");
        if (index > 0) { //suchwort mit * hintendran
          newrl = inf.getfileinfoByBeginningOfWord(word.substring(0, index), meta);
        }
        else {
          newrl = inf.getfileinfo(word, meta);
        }
	if (rulenum == Consts.AND_RULE) {
	    return ResultList.andresultlists(rl, newrl, andLevel);
	}
	if (rulenum == Consts.OR_RULE) {
	    return ResultList.orresultlists(rl, newrl);
	}
	if (rulenum == Consts.NOT_RULE) {
	    return ResultList.notresultlist(newrl, inf.maxfilenum);
	}
	return rl;
    }

    /**
     * Expand wildcards.
     * @param word list to be expanded
     * @return new word list
     */
    static StrList expandstars(IndexFile inf, StrList wlist) {
        StrList nlist = new StrList();
	if (wlist == null) {
	    return null;
	}
	String word = wlist.first();
        while (word != null) {
	    int index = word.indexOf("*");
	    if (index < 0) {
	        nlist.add(word);
            } else if(index == 0) { //suchwort ist *irgendwas
              boolean matchEnding = true;
              word = word.substring(1);
              int index2 = word.indexOf("*");
              if(index2>0){
                word = word.substring(0, index2);
                matchEnding = false;
              }
              boolean first = true;
              String found = null;
              do {
                  found = inf.getmatchword(word, first, matchEnding);
                  if (found != null) {
                      if (first) {
                          nlist.add("(");
                      } else {
                          nlist.add("or");
                      }
                      nlist.add(found);
                  } else {
                      if (first) {
                          nlist.add("thisisnotaword");
                      } else {
                          nlist.add(")");
                      }
                      break;
                  }
                  first = false;
		} while (found != null);
            } else {  //suchwort ist irgendwas*

              // ein suchwort irgendwas* bleibt zunaechst so in der wortliste und wird spaeter aufgeloest
              nlist.add(word);

              /* alte Variante bei der die Aufloesung hier erfolgt
	        word = word.substring(0, index);
		boolean first = true;
		String found = null;
		do {
		    found = inf.getmatchwordbyprefix(word, first);
		    if (found != null) {
		        if (first) {
			    nlist.add("(");
			} else {
			    nlist.add("or");
			}
			nlist.add(found);
		    } else {
		        if (first) {
			    nlist.add("thisisnotaword");
			} else {
			    nlist.add(")");
			}
			break;
		    }
		    first = false;
		} while (found != null);
              */
	    }
	    word = wlist.next();
	}
	return nlist;
    }

    /**
     * Perform keyword search.
     * @param inf index to be searched
     * @param words keywords to be searched
     * @param maxhits maximum number of hits
     * @param structure which meta information to be considered
     * @param base base URL or null, if not null result is HTML formatted
     * @return search result
     */
    static String search(IndexFile inf, String words, int maxhits,
			   int structure, URL base) {
	StrList wordlist = mkwlist(words);

	/* eliminate stopwords from wordlist */
	StrList newlist = new StrList();
	String t = wordlist.first();
	while (t != null) {
	    if (!inf.sh.isstopword(t) || isrule(t)) {
		newlist.add(t);
	    }
	    t = wordlist.next();
	}
	wordlist = newlist;
	if (wordlist.first() == null) {
	    if (base != null) {
	        return "<body>" +
		    "<h2>Error: words too common or empty.</h2>" +
		    "</body>";
	    }
	    return "err: search words too common or empty\n";
	}
	wordlist = fixnot(wordlist);

	if (Consts.DEBUG) {
	    System.out.print("*** Word list is: ");
	    String word = wordlist.first();
	    while (word != null) {
	        System.out.print(word + " ");
		word = wordlist.next();
	    }
	    System.out.println("");
	}

	wordlist = expandstars(inf, wordlist);
	if (Consts.DEBUG) {
	    System.out.print("*** Expanded word list is: ");
	    String word = wordlist.first();
	    while (word != null) {
	        System.out.print(word + " ");
		word = wordlist.next();
	    }
	    System.out.println("");
	}

	/* perform the search */
	SortInfo si = new SortInfo(words, base);
        SortResult sr = null;
	ResultList rl = parseterm(wordlist, false, 1, inf);
	for (Result r = rl.first; r != null; r = r.next) {
            if ((r.structure & structure) != 0) {
	        String fileinfo = inf.lookupfile(r.filenum);
		if (fileinfo != null) {
		    sr = SortResult.addSortResult(sr, r.rank, si, fileinfo);
		}
            }
        }

	StringBuffer sb = new StringBuffer();
	if (base == null) {
	    sb.append(inf.header);
	}
	if (sr != null) {
	    SortResult.printResults(sr, maxhits, si, sb);
	} else {
	    if (base != null) {
	        return "<body>" +
		     "<h2>Error: no hits.</h2>" +
		     "</body>";
	    }
	    sb.append("err: no hits\n");
	}
	return sb.toString();
    }


    static SortResultList getSortResult(IndexFile inf, String words, int maxhits,
                                int structure, URL base) {
      StrList wordlist = mkwlist(words);

      /* eliminate stopwords from wordlist */
      StrList newlist = new StrList();
      String t = wordlist.first();
      while (t != null) {
        if (!inf.sh.isstopword(t) || isrule(t)) {
          newlist.add(t);
        }
        t = wordlist.next();
      }
      wordlist = newlist;
      if (wordlist.first() == null) {
        //search words too common or empty
        return null;
      }
      wordlist = fixnot(wordlist);

      ///----------------------------------------
      if (Consts.DEBUG) {
        System.out.print("*** Word list is: ");
        String word = wordlist.first();
        while (word != null) {
          System.out.print(word + " ");
          word = wordlist.next();
        }
        System.out.println("");
      }
      ///----------------------------------------

      wordlist = expandstars(inf, wordlist);

      ///----------------------------------------
      if (Consts.DEBUG) {
        System.out.print("*** Expanded word list is: ");
        String word = wordlist.first();
        while (word != null) {
          System.out.print(word + " ");
          word = wordlist.next();
        }
        System.out.println("");
      }
      ///----------------------------------------

      /* perform the search */
      SortInfo si = new SortInfo(words, base);
      SortResult sr = null;
      ResultList rl = parseterm(wordlist, false, 1, inf);
      if(rl != null)
      {
        for (Result r = rl.first; r != null; r = r.next) {
          if ((r.structure & structure) != 0) {
            String fileinfo = inf.lookupfile(r.filenum);
            if (fileinfo != null) {
              sr = SortResult.addSortResult(sr, r.rank, si, fileinfo);
            }
          }
        }
      }
      else if (Consts.DEBUG)
      {
        System.out.println("*** ResultList null");
      }
      if (Consts.DEBUG) {
        System.out.println("*** Search done");
      }
      if(sr != null)
        return new SortResultList(sr);
      else
        return null;
    }
}
