/*
** Copyright (C) 1999-2001 Christian Werner
** modified by InterRed GmbH 03/2003: getResultInfos() added
**
** This program and library is free software; you can redistribute it and/or
** modify it under the terms of the GNU (Library) General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU (Library) General Public License for more details.
**
** You should have received a copy of the GNU (Library) General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package swishej;

import java.util.*;

/**
 * Sorting and formatting of a search result.
 * @see swishej.SortInfo
 */

class SortResult {
    /**
     * File information, usually filename, title, and size in bytes.
     */
    String fileinfo;

    /**
     * File information as array
     */
    String[] fileinfoArray;

    /**
     * Rank for sorting.
     */
    int rank;

    /**
     * Left and right result for sorting.
     */
    SortResult left, right;

    /**
     * Number of results left (or right).
     */
    int leftCount, rightCount;

    /**
     * Create a sort result entry.
     * @param rank rank for sorting
     * @param fileinfo file information of entry
     */
    SortResult(int rank, String fileinfo) {
	this.rank = rank;
	this.fileinfo = new String(fileinfo);
	this.left = null;
        this.leftCount = 0;
	this.right = null;
        this.rightCount = 0;
        this.fileinfoArray = splitfileinfo(fileinfo);
    }

    String[] getResultInfos() {
      return this.fileinfoArray;
    }

    /**
     * Add, i.e. insert a new entry to given sort result.
     * @param sr SortResult for accumulating sort results.
     * @param rank rank for sorting
     * @param si information for finding biggest rank.
     * @param fileinfo file information of entry
     * @return new SortResult entry
     */
    static SortResult addSortResult(SortResult sr, int rank, SortInfo si,
				    String fileinfo) {
	if (rank > si.bigrank) {
	    si.bigrank = rank;
	}
	if (sr == null) {
	    sr = new SortResult(rank, fileinfo);
	} else {
	    if (sr.rank < rank) {
	        sr.left = addSortResult(sr.left, rank, si, fileinfo);
                sr.leftCount++;
	    } else {
	        sr.right = addSortResult(sr.right, rank, si, fileinfo);
                sr.rightCount++;
	    }
	}
	return sr;
    }

    /**
     * Retrieve number of entries given sort result.
     * @param sr SortResult for counting
     * @return number of SortResults
     */
    static int countResults(SortResult sr) {
        if (sr == null) {
	    return 0;
	}
	return countResults(sr.right) + countResults(sr.left) + 1;
    }

    /**
     * Given rank print HTML representation of it.
     * @param rank rank of result
     * @return string with HTML representation of rank
     */
    static String printrank(int rank) {
        StringBuffer sb = new StringBuffer();
        double r = rank / 10.0;
	sb.append("<img src=");
	if (r >= 95.0) {
	    sb.append("star5.gif");
	} else if (r >= 80.0) {
	    sb.append("star4.gif");
	} else if (r >= 60.0) {
	    sb.append("star3.gif");
	} else if (r >= 30.0) {
	    sb.append("star2.gif");
	} else {
	    sb.append("star1.gif");
	}
	sb.append(" alt=\"" + r + "%\">");
	return sb.toString();
    }

    /**
     * Given file size print HTML representation of it.
     * @param size file size part of file information
     * @return string with HTML representation of size
     */
    static String printsize(String size) {
        Integer n;
        try {
	    n = Integer.valueOf(size);
	} catch (Exception e) {
	    return "unknown size";
	}
	if (n.intValue() >= 1024 * 1024) {
	    return "" + (n.intValue() / 1024 / 1024) + " MByte";
	} else if (n.intValue() >= 1024) {
	    return "" + (n.intValue() / 1024) + " kByte";
	}
	return "" + n + " Bytes";
    }

    /**
     * Split file information in pieces: 0. file name, 1. title, 2. file size
     * @param s file information
     * @return broken down String[] with file information
     */
    static String[] splitfileinfo(String s) {
        StringTokenizer st = new StringTokenizer(s, "\"");
	int count = 0;
	while (st.hasMoreTokens()) {
	    st.nextToken();
	    count++;
	}
	String result[] = new String[count];
	int i = 0;
	st = new StringTokenizer(s, "\"");
	while (st.hasMoreTokens()) {
	    result[i] = st.nextToken().trim();
	    i++;
	}
	return result;
    }

    /**
     * Recursively traverse sort result tree and format the result information.
     * @param sr current sort result
     * @param num scaling factor for printing rank
     * @param si SortInfo to be used for printing
     * @param sb StringBuffer which gets printed result representation appended
     */
    static void doprint(SortResult sr, double num, SortInfo si,
			StringBuffer sb) {
        if (sr != null) {
	    doprint(sr.left, num, si, sb);
	    int rank = (int) (sr.rank * num);
	    if (rank >= 999) {
	        rank = 1000;
	    } else if (rank <= 0) {
	        rank = 1;
	    }
	    if (si.maxhits != 0) {
	        if (si.htm) {
		    String fi[] = splitfileinfo(sr.fileinfo);
		    sb.append("<dt><a href=\"" + fi[0] + "\"><strong>" +
			      fi[1] +  "</strong></a> <small>(" +
			      printsize(fi[2]) + ")</small> " +
			      printrank(rank) + "\n<dd>");
		    String preview = null;
		    if (si.base != null) {
		        preview = Preview.preview(si.base, fi[0]);
		    }
		    if (preview != null) {
		        sb.append("<small>" + preview + "</small>\n");
		    }
		    sb.append("<br><br>\n");
		} else {
		    sb.append(rank + " " + sr.fileinfo + "\n");
		}
	        if (si.maxhits > 0) {
		    si.maxhits--;
		}
	    }
	    doprint(sr.right, num, si, sb);
	}
    }

    /**
     * Format the result information.
     * @param sr first sort result
     * @param max maximum number of hits to be printed
     * @param si SortInfo to be used for printing
     * @param sb StringBuffer which receives printed result representations
     */
    static void printResults(SortResult sr, int max, SortInfo si,
			     StringBuffer sb) {
        if (sr == null) {
	    if (si.htm) {
	        sb.append("<body><h2>No hits");
		if (si.words != null) {
		    sb.append(" for \"" + si.words + "\"");
		}
		sb.append(".</h2></body>");
	    } else {
	        sb.append("err: no results\n");
	    }
	    return;
	}
	double num = 1000.0;
	if (si.bigrank > 0) {
	    num = 1000.0 / si.bigrank;
	}
	int total = countResults(sr);
	if (si.htm) {
	    sb.append("<body><h2>" + total + " hits");
	    if (si.words != null) {
	        sb.append(" for \"" + si.words + "\"");
	    }
	    sb.append(".</h2>\n<dl>");
	} else {
	    sb.append("# Number of hits: " + total + "\n");
	}
	si.maxhits = max;
	doprint(sr, num, si, sb);
	if (si.htm) {
	    sb.append("</dl></body>\n");
	}
    }
}
